import io

import numpy as np
import PIL.Image
import requests


def load_image(url, size):
    r = requests.get(url)
    img = PIL.Image.open(io.BytesIO(r.content))
    img.thumbnail((40, 40), PIL.Image.ANTIALIAS)
    img = np.float32(img) / 255.0
    # premultiply RGB by Alpha
    img[..., :3] *= img[..., 3:]
    # pad to self.h, self.h
    diff = size - 40
    img = np.pad(img, ((diff // 2, diff // 2), (diff // 2, diff // 2), (0, 0)))
    return img


def load_emoji(emoji, size, code=None):
    if code is None:
        code = hex(ord(emoji))[2:].lower()
    url = (
        "https://github.com/googlefonts/noto-emoji/blob/main/png/128/emoji_u%s.png?raw=true"
        % code
    )
    return load_image(url, size)


def make_circle_masks(n, h, w, r=None):
    x = np.linspace(-1.0, 1.0, w)[None, None, :]
    y = np.linspace(-1.0, 1.0, h)[None, :, None]
    center = np.random.uniform(-0.5, 0.5, size=[2, n, 1, 1])
    if r is None:
        r = np.random.uniform(0.1, 0.4, size=[n, 1, 1])
    x, y = (x - center[0]) / r, (y - center[1]) / r
    mask = x * x + y * y < 1.0
    return mask.astype(float)



import matplotlib.pyplot as plt
import numpy as np
from einops import repeat

from jax_nca.utils import load_emoji


def to_alpha(x):
    return np.clip(x[:, :, :, 3:4], 0.0, 1.0)


def rgb(x, rgb=False):
    # assume rgb premultiplied by alpha
    if rgb:
        return np.clip(x[:, :, :, :3], 0.0, 1.0)
    rgb, a = x[:, :, :, :3], to_alpha(x)
    return np.clip(1.0 - a + rgb, 0.0, 1.0)


class ImageDataset:
    def __init__(self, emoji: str = None, img: np.array = None, img_size: int = 64):
        if img is None:
            img = load_emoji(emoji, img_size)
        self.rgb = img.shape[-1] == 3
        self.img_shape = img.shape
        self.img = np.expand_dims(img, 0)  # (b w h c)
        self.rgb_img = rgb(self.img, self.rgb)

    def get_batch(self, batch_size: int = 1):
        return repeat(
            self.img, "b w h c -> (b repeat) w h c", repeat=batch_size
        ), repeat(self.rgb_img, "b w h c -> (b repeat) w h c", repeat=batch_size)

    def visualize(self):
        _ = plt.imshow(self.rgb_img[0])
        plt.show()



import functools
import os
from collections.abc import Iterable
from datetime import datetime

import jax
import jax.numpy as jnp
import numpy as np
import optax
import pandas as pd
import tqdm
from flax.training import train_state  # Useful dataclass to keep train state
from tensorboardX import SummaryWriter

from jax_nca.utils import make_circle_masks


def get_tensorboard_logger(
    experiment_name: str, base_log_path: str = "tensorboard_logs"
):
    log_path = "{}/{}_{}".format(base_log_path, experiment_name, datetime.now())
    train_writer = SummaryWriter(log_path, flush_secs=10)
    full_log_path = os.path.join(os.getcwd(), log_path)
    print(
        "Follow tensorboard logs with: python -m tensorboard.main --logdir '{}'".format(
            full_log_path
        )
    )
    return train_writer


def create_train_state(rng, nca, learning_rate, shape):
    nca_seed = nca.create_seed(
        nca.num_hidden_channels, nca.num_target_channels, shape=shape[:-1], batch_size=1
    )
    """Creates initial `TrainState`."""
    params = nca.init(rng, nca_seed, rng)["params"]
    tx = optax.chain(
        # optax.clip_by_global_norm(10.0),
        optax.adam(learning_rate),
    )
    return train_state.TrainState.create(apply_fn=nca.apply, params=params, tx=tx)


def clip_grad_norm(grad):
    factor = 1.0 / (
        jnp.linalg.norm(jax.tree_util.tree_leaves(jax.tree_map(jnp.linalg.norm, grad)))
        + 1e-8
    )
    return jax.tree_map((lambda x: x * factor), grad)


@functools.partial(jax.jit, static_argnames=("apply_fn", "num_steps"))
def train_step(
    apply_fn, state, seeds: jnp.array, targets: jnp.array, num_steps: int, rng
):
    def mse_loss(pred, y):
        squared_diff = jnp.square(pred - y)
        return jnp.mean(squared_diff, axis=[-3, -2, -1])

    def loss_fn(params):
        def forward(carry, inp):
            carry = apply_fn({"params": params}, carry, rng)
            return carry, carry

        x, outs = jax.lax.scan(forward, seeds, None, length=num_steps)
        rgb, a = x[..., :3], jnp.clip(x[..., 3:4], 0.0, 1.0)
        rgb = jnp.clip(1.0 - a + rgb, 0.0, 1.0)

        outs = jnp.transpose(outs, [1, 0, 2, 3, 4])
        subset = outs[:, -8:]  # B 12 H W C
        return jnp.mean(
            jax.vmap(mse_loss)(subset[..., :4], jnp.expand_dims(targets, 1))
        ), (x, rgb)

    grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
    (loss, aux), grads = grad_fn(state.params)
    grads = clip_grad_norm(grads)
    updated, rgb = aux
    return state.apply_gradients(grads=grads), loss, grads, updated, rgb


class SamplePool:
    def __init__(self, max_size: int = 1000):
        self.max_size = max_size
        self.pool = [None] * max_size

    def __getitem__(self, idx):
        if isinstance(idx, Iterable):
            return [self.pool[i] for i in idx]
        return idx

    def __setitem__(self, idx, v):
        if isinstance(idx, Iterable):
            for i in range(len(idx)):
                index = idx[i]
                self.pool[index] = v[i]
        else:
            self.pool[idx] = v

    def sample(self, num_samples: int):
        indices = np.random.randint(0, self.max_size, num_samples)
        return self.__getitem__(indices), indices


def flatten(d):
    df = pd.json_normalize(d, sep="_")
    return df.to_dict(orient="records")[0]


class EmojiTrainer:
    def __init__(self, dataset, nca, pool_size: int = 1024, n_damage: int = 0):
        self.dataset = dataset
        self.img_shape = self.dataset.img_shape
        self.nca = nca
        self.pool_size = pool_size
        self.n_damage = n_damage
        self.state = None

    def train(
        self,
        num_epochs,
        batch_size: int = 8,
        seed: int = 10,
        lr: float = 0.001,
        min_steps: int = 64,
        max_steps: int = 96,
    ):
        pool = SamplePool(self.pool_size)

        writer = get_tensorboard_logger("EMOJITrainer")
        rng = jax.random.PRNGKey(seed)
        rng, init_rng = jax.random.split(rng)
        self.state = create_train_state(init_rng, self.nca, lr, self.dataset.img_shape)

        bar = tqdm.tqdm(np.arange(num_epochs))
        try:
            for i in bar:
                num_steps = int(np.random.randint(min_steps, max_steps))
                samples, indices = pool.sample(batch_size)
                for j in range(len(samples)):
                    if samples[j] is None:
                        samples[j] = self.nca.create_seed(
                            self.nca.num_hidden_channels,
                            self.nca.num_target_channels,
                            shape=self.img_shape[:-1],
                            batch_size=1,
                        )[0]
                samples[0] = self.nca.create_seed(
                    self.nca.num_hidden_channels,
                    self.nca.num_target_channels,
                    shape=self.img_shape[:-1],
                    batch_size=1,
                )[0]
                batch = np.stack(samples)
                if self.n_damage > 0:
                    damage = (
                        1.0
                        - make_circle_masks(
                            int(self.n_damage), self.img_shape[0], self.img_shape[1]
                        )[..., None]
                    )
                    batch[-self.n_damage :] *= damage

                batch = jnp.array(batch)
                targets, rgb_targets = self.dataset.get_batch(batch_size)
                targets = jnp.array(targets)

                self.state, loss, grads, outputs, rgb_outputs = train_step(
                    self.nca.apply,
                    self.state,
                    batch,
                    targets,
                    num_steps=num_steps,
                    rng=rng,
                )

                grad_dict = {k: dict(grads[k]) for k in grads.keys()}
                grad_dict = flatten(grad_dict)

                grad_dict = {
                    k: {kk: np.sum(vv).item() for kk, vv in v.items()}
                    for k, v in grad_dict.items()
                }
                grad_dict = flatten(grad_dict)

                pool[indices] = np.array(outputs)

                bar.set_description("Loss: {}".format(loss.item()))

                self.emit_metrics(
                    writer,
                    i,
                    batch,
                    rgb_outputs,
                    rgb_targets,
                    loss.item(),
                    metrics=grad_dict,
                )

            return self.state
        except Exception:
            return self.state

    def emit_metrics(
        self, train_writer, i: int, batch, outputs, targets, loss, metrics={}
    ):
        train_writer.add_scalar("loss", loss, i)
        # train_writer.add_scalar("log10(loss)", math.log10(loss), i)
        train_writer.add_images("batch", self.nca.to_rgb(batch), i, dataformats="NHWC")
        train_writer.add_images("outputs", outputs, i, dataformats="NHWC")
        train_writer.add_images("targets", targets, i, dataformats="NHWC")
        for k in metrics:
            train_writer.add_scalar(k, metrics[k], i)



import functools
from typing import Tuple

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax import serialization
from jax import lax


class SobelPerceptionNet(nn.Module):
    @nn.compact
    def __call__(self, x):
        # x shape - BHWC

        num_channels = x.shape[-1]

        # 2D sobel kernels - IOHW layout

        x_sobel_kernel = jnp.zeros(
            (num_channels, num_channels, 3, 3), dtype=jnp.float32
        )
        x_sobel_kernel += (
            jnp.array([[-1.0, 0.0, 1.0], [-2.0, 0.0, 2.0], [-1.0, 0.0, 1.0]])[
                jnp.newaxis, jnp.newaxis, :, :
            ]
            / 8.0
        )

        y_sobel_kernel = jnp.zeros(
            (num_channels, num_channels, 3, 3), dtype=jnp.float32
        )
        y_sobel_kernel += (
            jnp.array([[-1.0, -2.0, -1.0], [0.0, 0.0, 0.0], [1.0, 2.0, 1.0]])[
                jnp.newaxis, jnp.newaxis, :, :
            ]
            / 8.0
        )
        x = jnp.transpose(x, [0, 3, 1, 2])  # N C H W

        x_out = lax.conv(
            x,  # lhs = NCHW image tensor
            x_sobel_kernel,  # rhs = OIHW conv kernel tensor
            (1, 1),  # window strides
            "SAME",
        )  # padding mode

        y_out = lax.conv(
            x,  # lhs = NCHW image tensor
            y_sobel_kernel,  # rhs = OIHW conv kernel tensor
            (1, 1),  # window strides
            "SAME",
        )  # padding mode

        out = jnp.concatenate([x, x_out, y_out], axis=1)
        return jnp.transpose(out, [0, 2, 3, 1])  # N H W C


class UpdateNet(nn.Module):

    num_channels: int

    @nn.compact
    def __call__(self, x):
        update_layer_1 = nn.Conv(
            features=64, kernel_size=(1, 1), strides=1, padding="VALID"
        )
        update_layer_2 = nn.Conv(
            features=64, kernel_size=(1, 1), strides=1, padding="VALID"
        )
        update_layer_3 = nn.Conv(
            features=self.num_channels,
            kernel_size=(1, 1),
            strides=1,
            padding="VALID",
            kernel_init=jax.nn.initializers.zeros,
            use_bias=False,
        )
        x = update_layer_1(x)
        x = nn.relu(x)
        x = update_layer_2(x)
        x = nn.relu(x)
        x = update_layer_3(x)
        return x


class TrainablePerception(nn.Module):
    num_channels: int

    @nn.compact
    def __call__(self, x):
        out = nn.Conv(
            features=self.num_channels * 3,
            kernel_size=(3, 3),
            use_bias=False,
            feature_group_count=self.num_channels,
        )(x)
        return out


@functools.partial(jax.jit, static_argnames=("apply_fn", "num_steps"))
def nca_multi_step(apply_fn, params, current_state: jnp.array, rng, num_steps: int):
    def forward(carry, inp):
        carry = apply_fn({"params": params}, carry, rng)
        return carry, carry

    x, outs = jax.lax.scan(forward, current_state, None, length=num_steps)
    return x, outs


class NCA(nn.Module):
    num_hidden_channels: int
    num_target_channels: int = 3
    alpha_living_threshold: float = 0.1
    cell_fire_rate: float = 1.0
    trainable_perception: bool = False
    alpha: float = 1.0

    """
        num_hidden_channels: Number of hidden channels for each cell to use
        num_target_channels: Number of target channels to be used
        alpha_living_threshold: threshold to determine whether a cell lives or dies
        cell_fire_rate: probability that a cell receives an update per step
        trainable_perception: if true, instead of using sobel filters use a trainable conv net
        alpha: scalar value to be multiplied to updates
    """

    @classmethod
    def create_seed(
        cls,
        num_hidden_channels: int,
        num_target_channels: int = 3,
        shape: Tuple[int] = (48, 48),
        batch_size: int = 1,
    ):
        seed = np.zeros((batch_size, *shape, num_hidden_channels + 3 + 1))
        w, h = seed.shape[1], seed.shape[2]
        seed[:, w // 2, h // 2, 3:] = 1.0
        return seed

    def setup(self):
        num_channels = 3 + self.num_hidden_channels + 1
        if self.trainable_perception:
            self.perception = TrainablePerception(num_channels)
        else:
            self.perception = SobelPerceptionNet()
        self.update_net = UpdateNet(num_channels)

    def alive(self, x, alpha_living_threshold: float):
        return (
            nn.max_pool(
                x[..., 3:4], window_shape=(3, 3), strides=(1, 1), padding="SAME"
            )
            > alpha_living_threshold
        )

    def get_stochastic_update_mask(self, x, rng, cell_fire_rate: float = 1.0):
        return jnp.array(np.random.uniform(size=x[..., :1].shape) <= cell_fire_rate)

    def __call__(self, x, rng):
        pre_life_mask = self.alive(x, self.alpha_living_threshold)

        perception_out = self.perception(x)
        update = self.alpha * jnp.reshape(self.update_net(perception_out), x.shape)

        if self.cell_fire_rate >= 1.0:
            stochastic_update_mask = self.get_stochastic_update_mask(
                x, rng, self.cell_fire_rate
            ).astype(float)
            x = x + update * stochastic_update_mask
        else:
            x = x + update

        post_life_mask = self.alive(x, self.alpha_living_threshold)

        life_mask = pre_life_mask & post_life_mask
        life_mask = life_mask.astype(float)

        return x * life_mask

    def save(self, params, path: str):
        bytes_output = serialization.to_bytes(params)
        with open(path, "wb") as f:
            f.write(bytes_output)

    def load(self, path: str):
        nca_seed = self.create_seed(
            self.num_hidden_channels, self.num_target_channels, batch_size=1
        )
        rng = jax.random.PRNGKey(0)
        init_params = self.init(rng, nca_seed, rng)["params"]
        with open(path, "rb") as f:
            bytes_output = f.read()
        return serialization.from_bytes(init_params, bytes_output)

    def multi_step(self, params, current_state: jnp.array, rng, num_steps: int = 2):
        return nca_multi_step(self.apply, params, current_state, rng, num_steps)

    def to_rgb(self, x: jnp.array):
        rgb, a = x[..., :3], jnp.clip(x[..., 3:4], 0.0, 1.0)
        rgb = jnp.clip(1.0 - a + rgb, 0.0, 1.0)
        return rgb
